// =============================================================================
// Copyright (C) MACNICA inc.
// =============================================================================
// FILE NAME      : i2c_master_top.v 
// DEPARTMENT     : MACNICA Strategic Technology Group
// AUTHOR         : Youhei Nagano
// AUTHOR'S EMAIL : Nagano-y@macnica.co.jp
// =============================================================================
// KEYWORDS :  PRJ-000114
// =============================================================================
// PURPOSE  :  Scan Conversion IP Reference Design
// =============================================================================
// COMMENTS : Registered by Nagano , 100511
// =============================================================================

module i2c_master_top (
    clk             ,
    rst_n           ,
    srst            ,

    scl_i           ,
    scl_o           ,
    scl_en_o        ,
    sda_i           ,
    sda_o           ,
    sda_en_o        ,

    control_reg_i   ,
    dev_sel_i       ,
    w_r_adr_i       ,
    b_size_i        ,
    div_i           ,
    start_i         ,

    busy_o          ,
    error_o         ,
    rem_num_o       ,

    rd_data_o       ,
    rd_d_valid_o    ,
    wt_data_i       ,
    wt_d_req_o      
    );

/****************************************************************/
/*****                  parameter definition                *****/
/****************************************************************/

    parameter   DIV_COUNT_BIT_W     = 8 ;


    `define     SCL_MAX     4'b1000
    `define     SCL_MAX_1M  4'b0111
    `define     SCL_ZERO    4'b0000
    `define     SCL_ONE     4'b0001
    `define     CLK_ZERO    {DIV_COUNT_BIT_W{1'b0}}
    `define     B_CNT_ONE   {{23{1'b0}},1'b1}
    `define     START_BYTE  8'b00000001
    `define     WT_BIT      1'b0
    `define     RD_BIT      1'b1
//  main state 
    `define     M_IDLE          2'b00
    `define     M_RANDAM_RD     2'b01
    `define     M_CURRENT_RD    2'b10
    `define     M_RANDAM_WT     2'b11
//  sub state
    `define     S_IDLE          4'b0000
    `define     S_START         4'b0001
    `define     S_START_BYTE    4'b0010
    `define     S_START_R       4'b0011
    `define     S_DEVSEL_H_W    4'b0100
    `define     S_DEVSEL_L_W    4'b0101
    `define     S_DEVSEL_R      4'b0110
    `define     S_ADDRESS1      4'b0111
    `define     S_ADDRESS2      4'b1000
    `define     S_ADDRESS3      4'b1001
    `define     S_DATA_RD       4'b1010
    `define     S_DATA_WT       4'b1011
    `define     S_STOP          4'b1100

/****************************************************************/
/*****                  port declaration                    *****/
/****************************************************************/

    input                               clk             ;
    input                               rst_n           ;
    input                               srst            ;

    input                               scl_i           ;
    output                              scl_o           ;
    output                              scl_en_o        ;
    input                               sda_i           ;
    output                              sda_o           ;
    output                              sda_en_o        ;

    input   [                 6 : 0 ]   control_reg_i   ;
    input   [                 9 : 0 ]   dev_sel_i       ;
    input   [                23 : 0 ]   w_r_adr_i       ;
    input   [                23 : 0 ]   b_size_i        ;
    input   [ DIV_COUNT_BIT_W-1 : 0 ]   div_i           ;
    input                               start_i         ;

    output                              busy_o          ;
    output                              error_o         ;
    output  [                23 : 0 ]   rem_num_o       ;

    output  [                 7 : 0 ]   rd_data_o       ;
    output                              rd_d_valid_o    ;
    input   [                 7 : 0 ]   wt_data_i       ;
    output                              wt_d_req_o      ;

/****************************************************************/
/*****                  reg/wire declaration                *****/
/****************************************************************/

    reg     [                 1 : 0 ]   main_state      ;
    reg     [                 3 : 0 ]   sub_state       ;

    reg                                 rand_rd         ;   //  randam read control register
    reg                                 cur_rd          ;   //  current read control register
    reg                                 rand_wt         ;   //  write contorol register
    reg                                 dev_sel_size    ;   //  0:address=7bit 1:address=10bit
    reg                                 s_byte          ;   //  START byte  0:not output 1:output
    reg     [                 1 : 0 ]   w_r_adr_bit     ;   //  00: notADR 01:8bitAddress 10:16bitAddress 11:24bitAddress
    reg     [                 9 : 0 ]   dev_sel         ;
    reg     [                23 : 0 ]   w_r_adr         ;
    reg     [ DIV_COUNT_BIT_W-1 : 0 ]   div             ;   //
    wire    [ DIV_COUNT_BIT_W-1 : 0 ]   div_max         ;   //
    wire    [ DIV_COUNT_BIT_W-1 : 0 ]   div_half        ;   //
    wire    [ DIV_COUNT_BIT_W-1 : 0 ]   div_quarter     ;   //
    wire    [ DIV_COUNT_BIT_W-1 : 0 ]   div_one_8th     ;   //
    wire    [ DIV_COUNT_BIT_W-1 : 0 ]   div_quarter_3m  ;   //
    wire    [ DIV_COUNT_BIT_W-1 : 0 ]   scl_low         ;   //
    wire    [ DIV_COUNT_BIT_W-1 : 0 ]   scl_high        ;   //

    reg     [ DIV_COUNT_BIT_W-1 : 0 ]   clk_count       ;   //  
    reg     [                 3 : 0 ]   scl_count       ;   //  scl counter
    reg     [                23 : 0 ]   b_count         ;   //  burst size
    reg     [                 7 : 0 ]   out_data        ;   //  output data
    reg                                 out_d_req       ;   //  sda output data req

    reg                                 i2c_start       ;   //  
    reg                                 start_i_lat     ;   //  start_i latch signal

    reg                                 scl             ;   //
    reg                                 scl_en          ;
    reg                                 scl_lat1        ;
    reg                                 scl_lat2        ;
    reg                                 sda             ;
    reg                                 sda_en          ;
    reg                                 busy            ;   //
    reg                                 error           ;
    reg     [                 7 : 0 ]   rd_data         ;
    reg                                 rd_d_valid      ;
    reg                                 wt_d_req        ;
    wire                                next_state      ;
    wire                                nack            ;


//--------------------
//--  input signal  --
//--------------------
//--  I2C bus
    assign nack = sda_i ;

    always @ (posedge clk or negedge rst_n ) begin
        if ( ~rst_n ) begin
            scl_lat1 <= 1'b0 ;
            scl_lat2 <= 1'b0 ;
        end else if ( srst ) begin
            scl_lat1 <= 1'b0 ;
            scl_lat2 <= 1'b0 ;
        end else begin
            scl_lat1 <= scl_i ;
            scl_lat2 <= scl_lat1 ;
        end
    end

//--  Local I/F
    always @ ( posedge clk or negedge rst_n ) begin
        if ( ~rst_n ) begin
            rand_rd         <= 1'b0 ;
            cur_rd          <= 1'b0 ;
            rand_wt         <= 1'b0 ;
            dev_sel_size    <= 1'b0 ;
            dev_sel         <= { 10{1'b0} } ;
            w_r_adr         <= { 24{1'b0} } ;
            div             <= { DIV_COUNT_BIT_W{1'b0} } ;
        end else if ( srst ) begin
            rand_rd         <= 1'b0 ;
            cur_rd          <= 1'b0 ;
            rand_wt         <= 1'b0 ;
            dev_sel_size    <= 1'b0 ;
            dev_sel         <= { 10{1'b0} } ;
            w_r_adr         <= { 24{1'b0} } ;
            div             <= { DIV_COUNT_BIT_W{1'b0} } ;
        end else if ( start_i & ~busy ) begin
            rand_rd         <= control_reg_i[0] ;
            cur_rd          <= control_reg_i[1] ;
            rand_wt         <= control_reg_i[2] ;
            dev_sel_size    <= control_reg_i[3] ;
            dev_sel         <= dev_sel_i ;
            w_r_adr         <= w_r_adr_i ;
            div             <= div_i ;
        end
    end

    always @ ( posedge clk or negedge rst_n ) begin
        if ( ~rst_n ) begin
            w_r_adr_bit     <= 2'b00 ;
        end else if ( srst ) begin
            w_r_adr_bit     <= 2'b00 ;
        end else if ( start_i & ~busy ) begin
            w_r_adr_bit     <= control_reg_i[ 5 : 4 ] ;
        end else if ( i2c_start & cur_rd ) begin
            w_r_adr_bit     <= 2'b00 ;
        end
    end

    always @ ( posedge clk or negedge rst_n ) begin
        if ( ~rst_n ) begin
            s_byte <= 1'b0 ;
        end else if ( srst ) begin
            s_byte <= 1'b0 ;
        end else if ( start_i & ~busy ) begin
            s_byte <= control_reg_i[6] ;
        end else if ( sub_state==`S_START_R & clk_count==`CLK_ZERO  & scl_count==`SCL_ONE ) begin
            s_byte <= 1'b0 ;
        end
    end

    assign div_max          = div - 1'b1 ;
    assign div_half         = { 1'b0 , div[ DIV_COUNT_BIT_W-1 : 1 ] } ;
    assign div_quarter      = { 2'b00 , div[ DIV_COUNT_BIT_W-1 : 2 ] } ;
    assign div_one_8th      = { 3'b000 , div[ DIV_COUNT_BIT_W-1 : 3 ] } ;
    assign div_quarter_3m   = div_quarter - 2'b11 ;

    assign scl_low  = div_half + div_quarter + div_one_8th ;
    assign scl_high = div_quarter ;

//---------------------
//--  output signal  --
//---------------------
//--  I2C bus
    assign scl_o = scl ;
    always @ ( posedge clk or negedge rst_n ) begin
        if ( ~rst_n ) begin
            scl <= 1'b1 ;
        end else if ( srst ) begin
            scl <= 1'b1 ;
        end else if ( ( sub_state!=`S_IDLE & sub_state!=`S_START & sub_state!=`S_STOP & sub_state!=`S_START_R ) |
                    ( sub_state==`S_STOP & scl_count==`SCL_ZERO ) |
                    ( sub_state==`S_START_R & scl_count==`SCL_ZERO ) ) begin
            if ( clk_count==scl_low ) begin
                scl <= 1'b0 ;
            end else if ( clk_count==scl_high ) begin
                scl <= 1'b1 ;
            end
        end
    end

    assign scl_en_o = scl_en ;
    always @ ( posedge clk or negedge rst_n ) begin
        if ( ~rst_n ) begin
            scl_en <= 1'b0 ;
        end else if ( srst ) begin
            scl_en <= 1'b0 ;
        end else if ( sub_state!=`S_IDLE ) begin
            scl_en <= 1'b1 ;
        end else begin
            scl_en <= 1'b0 ;
        end
    end

    assign sda_o = sda ;
    always @ ( posedge clk or negedge rst_n ) begin
        if ( ~rst_n ) begin
            sda <= 1'b1 ;
        end else if ( srst ) begin
            sda <= 1'b1 ;
        end else if ( clk_count==div_half ) begin
            if ( sub_state==`S_START |
                 ( sub_state==`S_START_R & scl_count==`SCL_ONE ) |
                 ( sub_state==`S_STOP & scl_count==`SCL_ZERO ) ) begin
                sda <= 1'b0 ;
            end else if ( sub_state==`S_STOP |
                    ( sub_state==`S_START_R & scl_count==`SCL_ZERO ) |
                    ( b_count==`B_CNT_ONE & scl_count==`SCL_MAX ) ) begin
                sda <= 1'b1 ;
            end else begin
                sda <= out_data[7] ;
            end
        end
    end

    assign sda_en_o = sda_en ;
    always @ ( posedge clk or negedge rst_n ) begin
        if ( ~rst_n ) begin
            sda_en <= 1'b0 ;
        end else if ( srst ) begin
            sda_en <= 1'b0 ;
        end else if ( ( sub_state==`S_IDLE ) |
                      ( sub_state==`S_STOP & clk_count==`CLK_ZERO & scl_count==`SCL_ONE ) ) begin
            sda_en <= 1'b0 ;
        end else if ( sub_state==`S_START ) begin
            sda_en <= 1'b1 ;
        end else if ( clk_count==div_half & scl_count==`SCL_ZERO ) begin
            if ( sub_state==`S_DATA_RD ) begin
                sda_en <= 1'b0 ;
            end else begin
                sda_en <= 1'b1 ;
            end
        end else if ( clk_count==div_half & scl_count==`SCL_MAX ) begin
            if ( sub_state==`S_DATA_RD ) begin
                sda_en <= 1'b1 ;
            end else begin
                sda_en <= 1'b0 ;
            end
        end
    end

//--  Local I/F
    assign rd_data_o = rd_data ;
    always @ ( posedge clk or negedge rst_n ) begin
        if ( ~rst_n ) begin
            rd_data <= 8'b00000000 ;
        end else if ( srst ) begin
            rd_data <= 8'b00000000 ;
        end else if ( sub_state==`S_DATA_RD & clk_count==`CLK_ZERO & scl_count!=`SCL_MAX ) begin
            rd_data <= { rd_data[ 6 : 0 ] , sda_i } ;
        end
    end

    assign rd_d_valid_o = rd_d_valid ;
    always @ ( posedge clk or negedge rst_n ) begin
        if ( ~rst_n ) begin
            rd_d_valid <= 1'b0 ;
        end else if ( srst ) begin
            rd_d_valid <= 1'b0 ;
        end else if ( sub_state==`S_DATA_RD & clk_count==`CLK_ZERO & scl_count==`SCL_MAX ) begin
            rd_d_valid <= 1'b1 ;
        end else begin
            rd_d_valid <= 1'b0 ;
        end
    end

    assign wt_d_req_o = wt_d_req ;
    always @ ( posedge clk or negedge rst_n ) begin
        if ( ~rst_n ) begin
            wt_d_req <= 1'b0 ;
        end else if ( srst ) begin
            wt_d_req <= 1'b0 ;
        end else if ( sub_state==`S_DATA_WT & clk_count==div_max & scl_count==`SCL_ZERO ) begin
            wt_d_req <= 1'b1 ;
        end else begin
            wt_d_req <= 1'b0 ;
        end
    end

    assign busy_o = busy ;
    always @ ( posedge clk or negedge rst_n ) begin
        if ( ~rst_n ) begin
            busy <= 1'b0 ;
        end else if ( srst ) begin
            busy <= 1'b0 ;
        end else if ( i2c_start ) begin
            busy <= 1'b1 ;
        end else if ( main_state==`M_IDLE ) begin
            busy <= 1'b0 ;
        end
    end

    assign error_o = error ;
    always @ ( posedge clk or negedge rst_n ) begin
        if ( ~rst_n ) begin
            error <= 1'b0 ;
        end else if ( srst ) begin
            error <= 1'b0 ;
        end else if ( sub_state!=`S_DATA_RD & sub_state!=`S_START_BYTE & 
                      clk_count==`CLK_ZERO & scl_count==`SCL_MAX & nack ) begin
            error <= 1'b1 ;
        end else if ( i2c_start ) begin
            error <= 1'b0 ;
        end
    end

    assign rem_num_o = b_count ;

//--  out put sda data 
    always @ ( posedge clk or negedge rst_n ) begin
        if ( ~rst_n ) begin
            out_data <= 8'b00000000 ;
        end else if ( srst ) begin
            out_data <= 8'b00000000 ;
        end else if ( out_d_req ) begin
            if ( scl_count==`SCL_ZERO ) begin
                case ( sub_state )
                    `S_START_BYTE : begin
                        out_data <= `START_BYTE ;
                    end
                    
                    `S_DEVSEL_H_W : begin
                        out_data <= { 5'b11110 , dev_sel[9] , dev_sel[8] , `WT_BIT } ;
                    end
                    
                    `S_DEVSEL_L_W : begin
                        if ( dev_sel_size==1'b0 ) begin
                            out_data <= { dev_sel[ 6 : 0] , `WT_BIT } ;
                        end else begin
                            out_data <= dev_sel [7 : 0 ];
                        end
                    end
                    
                    `S_DEVSEL_R : begin
                        if ( ~dev_sel_size ) begin
                            out_data <= { dev_sel[ 6 : 0] , `RD_BIT } ;
                        end else begin
                            out_data <= { 5'b11110 , dev_sel[9] , dev_sel[8] , `RD_BIT } ;
                        end 
                    end
                    
                    `S_ADDRESS1 : begin
                        out_data <= w_r_adr[ 23 : 16 ] ;
                    end
                    
                    `S_ADDRESS2 : begin
                        out_data <= w_r_adr[ 15 : 8 ] ;
                    end
                    
                    `S_ADDRESS3 : begin
                        out_data <= w_r_adr[ 7 : 0 ] ;
                    end
                    
                    `S_DATA_WT : begin
                        out_data <= wt_data_i ;
                    end
                    
                    default : begin
                        out_data <= 8'bXXXXXXXX ;
                    end
                endcase
            end else begin
                out_data <= { out_data[ 6 : 0 ] , 1'b0 } ;
            end 
        end 
    end

    always @ ( posedge clk or negedge rst_n ) begin
        if ( ~rst_n ) begin
            out_d_req <= 1'b0 ;
        end else if ( srst ) begin
            out_d_req <= 1'b0 ;
        end else if ( clk_count==div_max ) begin
            out_d_req <= 1'b1 ;
        end else begin
            out_d_req <= 1'b0 ;
        end
    end

//--------------------
//--  start signal  --
//--------------------
    always @ ( posedge clk or negedge rst_n ) begin
        if ( ~rst_n ) begin
            i2c_start <= 1'b0 ;
        end else if ( srst ) begin
            i2c_start <= 1'b0 ;
        end else if ( start_i & (~start_i_lat) & (~busy) ) begin
            i2c_start <= 1'b1 ;
        end else begin
            i2c_start <= 1'b0 ;
        end
    end

    always @ ( posedge clk or negedge rst_n ) begin
        if ( ~rst_n ) begin
            start_i_lat <= 1'b0 ;
        end else if ( srst ) begin
            start_i_lat <= 1'b0 ;
        end else begin
            start_i_lat <= start_i ;
        end
    end

//---------------------
//--  state machine  --
//---------------------
//--
    always @ ( posedge clk or negedge rst_n ) begin
        if ( ~rst_n ) begin
            main_state <= `M_IDLE ;
        end else if ( srst ) begin
            main_state <= `M_IDLE ;
        end else begin
            case ( main_state )
                `M_IDLE : begin
                    if ( i2c_start ) begin
                        if ( rand_rd ) begin
                            if ( dev_sel_size==1'b0 & w_r_adr_bit==2'b00 ) begin
                                main_state <= `M_CURRENT_RD ;
                            end else begin
                                main_state <= `M_RANDAM_RD ;
                            end
                        end else if ( cur_rd ) begin
                            if ( dev_sel_size==1'b0 ) begin
                                main_state <= `M_CURRENT_RD ;
                            end else begin
                                main_state <= `M_RANDAM_RD ;
                            end
                        end else if ( rand_wt ) begin
                            main_state <= `M_RANDAM_WT ;
                        end
                    end
                end

                default : begin
                    if ( sub_state==`S_STOP & clk_count==`CLK_ZERO & scl_count==`SCL_ONE ) begin
                        main_state <= `M_IDLE ;
                    end
                end
            endcase
        end
    end


    always @ ( posedge clk or negedge rst_n ) begin
        if ( ~rst_n ) begin
            sub_state <= `S_IDLE ;
        end else if ( srst ) begin
            sub_state <= `S_IDLE ;
        end else begin
            case ( main_state )
                `M_IDLE : begin
                    sub_state <= sub_state ;
                end

//              --  Randam Read Sequence
                `M_RANDAM_RD : begin
                    case ( sub_state )
                        `S_IDLE : begin
                            sub_state <= `S_START ;
                        end
                        
                        `S_START : begin
                            if ( clk_count==`CLK_ZERO ) begin
                                if( s_byte ) begin
                                    sub_state <= `S_START_BYTE ;
                                end else if ( dev_sel_size ) begin
                                    sub_state <= `S_DEVSEL_H_W ;
                                end else begin
                                    sub_state <= `S_DEVSEL_L_W ;
                                end
                            end
                        end
                        
                        `S_START_BYTE : begin
                            if ( next_state ) begin
                                sub_state <= `S_START_R ;
                            end
                        end
                        
                        `S_DEVSEL_H_W : begin
                            if ( next_state ) begin
                                if ( nack ) begin
                                    sub_state <= `S_STOP ;
                                end else begin
                                    sub_state <= `S_DEVSEL_L_W ;
                                end 
                            end 
                        end
                        
                        `S_DEVSEL_L_W : begin
                            if ( next_state ) begin
                                if ( nack ) begin
                                    sub_state <= `S_STOP ;
                                end else if ( w_r_adr_bit==2'b11 ) begin
                                    sub_state <= `S_ADDRESS1 ;
                                end else if ( w_r_adr_bit==2'b10 ) begin
                                    sub_state <= `S_ADDRESS2 ;
                                end else if ( w_r_adr_bit==2'b01 ) begin
                                    sub_state <= `S_ADDRESS3 ;
                                end else if ( w_r_adr_bit==2'b00 ) begin
                                    sub_state <= `S_START_R ;
                                end 
                            end
                        end

                        `S_ADDRESS1 : begin
                            if ( next_state ) begin
                                if ( nack ) begin
                                    sub_state <= `S_STOP ;
                                end else begin
                                    sub_state <= `S_ADDRESS2 ;
                                end
                            end
                        end

                        `S_ADDRESS2 : begin
                            if ( next_state ) begin
                                if ( nack ) begin
                                    sub_state <= `S_STOP ;
                                end else begin
                                    sub_state <= `S_ADDRESS3;
                                end 
                            end 
                        end
                        
                        `S_ADDRESS3 : begin
                            if ( next_state ) begin
                                if ( nack ) begin
                                    sub_state <= `S_STOP ;
                                end else begin
                                    sub_state <= `S_START_R ;
                                end 
                            end 
                        end
                        
                        `S_START_R : begin
                            if ( clk_count==`CLK_ZERO  & scl_count==`SCL_ONE ) begin
                                if ( s_byte ) begin
                                    if ( dev_sel_size ) begin
                                        sub_state <= `S_DEVSEL_H_W ;
                                    end else begin
                                        sub_state <= `S_DEVSEL_L_W ;
                                    end
                                end else begin
                                    sub_state <= `S_DEVSEL_R ;
                                end
//                                sub_state <= `S_DEVSEL_R ;
                            end
                        end
                        
                        `S_DEVSEL_R : begin
                            if ( next_state ) begin
                                if ( nack ) begin
                                    sub_state <= `S_STOP ;
                                end else begin
                                    sub_state <= `S_DATA_RD ;
                                end 
                            end 
                        end

                        `S_DATA_RD : begin
                            if ( next_state ) begin
                                if ( b_count==`B_CNT_ONE ) begin
                                    sub_state <= `S_STOP ;
                                end 
                            end 
                        end 
                        
                        `S_STOP  : begin
                            if ( clk_count==`CLK_ZERO & scl_count==`SCL_ONE ) begin
                                sub_state <= `S_IDLE ;
                            end 
                        end 

                        default : begin
                            sub_state <= sub_state ;
                        end 

                    endcase
                end
                
//              --  Current Read Sequence
                `M_CURRENT_RD : begin
                    case ( sub_state )
                        `S_IDLE : begin
                            sub_state <= `S_START ;
                        end
                        
                        `S_START : begin
                            if ( clk_count==`CLK_ZERO ) begin
                                if ( s_byte ) begin
                                    sub_state <= `S_START_BYTE ;
                                end else begin
                                    sub_state <= `S_DEVSEL_R ;
                                end
                            end
                        end
                        
                        `S_START_BYTE : begin
                            if ( next_state ) begin
                                sub_state <= `S_START_R ;
                            end
                        end

                        `S_START_R : begin
                            if ( clk_count==`CLK_ZERO  & scl_count==`SCL_ONE ) begin
                                sub_state <= `S_DEVSEL_R ;
                            end
                        end
                        
                        `S_DEVSEL_R : begin
                            if ( next_state ) begin
                                if ( nack ) begin
                                    sub_state <= `S_STOP ;
                                end else begin
                                    sub_state <= `S_DATA_RD ;
                                end 
                            end
                        end 

                        `S_DATA_RD : begin
                            if ( next_state ) begin
                                if ( b_count==`B_CNT_ONE ) begin
                                    sub_state <= `S_STOP ;
                                end 
                            end 
                        end
                        
                        `S_STOP : begin
                            if ( clk_count==`CLK_ZERO & scl_count==`SCL_ONE ) begin
                                sub_state <= `S_IDLE ;
                            end
                        end

                        default : begin
                            sub_state <= sub_state ;
                        end

                    endcase
                end
                
//              --  Randam Write Sequence
                `M_RANDAM_WT : begin
                    case ( sub_state )
                        `S_IDLE : begin
                            sub_state <= `S_START ;
                        end
                        
                        `S_START : begin
                            if ( clk_count==`CLK_ZERO ) begin
                                if( s_byte ) begin
                                    sub_state <= `S_START_BYTE ;
                                end else if ( dev_sel_size ) begin
                                    sub_state <= `S_DEVSEL_H_W ;
                                end else begin
                                    sub_state <= `S_DEVSEL_L_W ;
                                end
                            end 
                        end
                      
                        `S_START_BYTE : begin
                            if ( next_state ) begin
                                sub_state <= `S_START_R ;
                            end
                        end
                        
                        `S_START_R : begin
                            if ( clk_count==`CLK_ZERO  & scl_count==`SCL_ONE ) begin
                                if ( dev_sel_size ) begin
                                    sub_state <= `S_DEVSEL_H_W ;
                                end else begin
                                    sub_state <= `S_DEVSEL_L_W ;
                                end
                            end
                        end
                        
                        `S_DEVSEL_H_W : begin
                            if ( next_state ) begin
                                if ( nack ) begin
                                    sub_state <= `S_STOP ;
                                end else begin
                                    sub_state <= `S_DEVSEL_L_W ;
                                end 
                            end 
                        end
                        
                        `S_DEVSEL_L_W : begin
                            if ( next_state ) begin
                                if ( nack ) begin
                                    sub_state <= `S_STOP ;
                                end else if ( w_r_adr_bit==2'b11 ) begin
                                    sub_state <= `S_ADDRESS1 ;
                                end else if ( w_r_adr_bit==2'b10 ) begin
                                    sub_state <= `S_ADDRESS2 ;
                                end else if ( w_r_adr_bit==2'b01 ) begin
                                    sub_state <= `S_ADDRESS3 ;
                                end else if ( w_r_adr_bit==2'b00 ) begin
                                    sub_state <= `S_DATA_WT ;
                                end 
                            end 
                        end

                        `S_ADDRESS1 : begin
                            if ( next_state ) begin
                                if ( nack ) begin
                                    sub_state <= `S_STOP ;
                                end else begin
                                    sub_state <= `S_ADDRESS2 ;
                                end 
                            end 
                        end 

                        `S_ADDRESS2 : begin
                            if ( next_state ) begin
                                if ( nack ) begin
                                    sub_state <= `S_STOP ;
                                end else begin
                                    sub_state <= `S_ADDRESS3 ;
                                end 
                            end 
                        end
                        
                        `S_ADDRESS3 : begin
                            if ( next_state ) begin
                                if ( nack ) begin
                                    sub_state <= `S_STOP ;
                                end else begin
                                    sub_state <= `S_DATA_WT ;
                                end 
                            end 
                        end 

                        `S_DATA_WT : begin
                            if ( next_state ) begin
                                if ( nack ) begin
                                    sub_state <= `S_STOP ;
                                end else if ( b_count==`B_CNT_ONE ) begin
                                    sub_state <= `S_STOP ;
                                end 
                            end 
                        end
                        
                        `S_STOP : begin
                            if ( clk_count==`CLK_ZERO & scl_count==`SCL_ONE ) begin
                                sub_state <= `S_IDLE ;
                            end 
                        end

                        default : begin
                            sub_state <= sub_state ;
                        end

                    endcase
                end
                
                default : begin
                    sub_state <= sub_state ;
                end
            endcase
        end
    end
    
    assign next_state = ( clk_count==`CLK_ZERO & scl_count==`SCL_MAX ) ? 1'b1 : 1'b0 ;


//---------------
//--  counter  --
//---------------
//--  clk counter
    always @ ( posedge clk or negedge rst_n ) begin
        if ( ~rst_n ) begin
            clk_count <=  { DIV_COUNT_BIT_W{1'b0} } ;
        end else if ( srst ) begin
            clk_count <=  { DIV_COUNT_BIT_W{1'b0} } ;
        end else if ( i2c_start ) begin
            clk_count <= div_max ;
        end else if ( sub_state!=`S_IDLE ) begin
            if ( clk_count==`CLK_ZERO ) begin
                clk_count <= div_max ;
            end else if ( clk_count==div_quarter_3m & scl_lat2==1'b0 ) begin
                clk_count <= clk_count ;
            end else begin
                clk_count <= clk_count - 1'b1 ;
            end
        end
    end

//--  scl counter
    always @ ( posedge clk or negedge rst_n ) begin
        if ( ~rst_n ) begin
            scl_count <= 4'b0000 ;
        end else if ( srst ) begin
            scl_count <= 4'b0000 ;
        end else if ( ( sub_state!=`S_IDLE & sub_state!=`S_START ) & clk_count==`CLK_ZERO ) begin
            if ( scl_count==`SCL_MAX | 
                 ( ( sub_state==`S_STOP | sub_state==`S_START_R ) & scl_count==`SCL_ONE ) ) begin
                scl_count <= 4'b0000 ;
            end else begin
                scl_count <= scl_count + 1'b1 ;
            end
        end
    end

//--  data burst counter
    always @ ( posedge clk or negedge rst_n ) begin
        if ( ~rst_n ) begin
            b_count <= { 24{1'b0} } ;
        end else if ( srst ) begin
            b_count <= { 24{1'b0} } ;
        end else if ( start_i & (~busy) ) begin
            b_count <= b_size_i ;
        end else if ( ( sub_state==`S_DATA_RD | ( sub_state==`S_DATA_WT & (~nack) ) ) &
                      clk_count==`CLK_ZERO & scl_count==`SCL_MAX ) begin
            b_count <= b_count - 1'b1 ;
        end
    end


endmodule
