// =============================================================================
// Copyright (C) MACNICA inc.
// =============================================================================
// FILE NAME	: avb_pkt_asm_avs2vsi.v
// DEPARTMENT	: MACNICA Strategic Technology Group
// AUTHOR		: Ryosuke Yamamoto
// DATE 		: 03/05/2014
// =============================================================================
// DESCRIPTION	:
//		
// =============================================================================
// REVISION HISTORY :
//		v1.0		: Initial Release
//
// =============================================================================
// PARAMETERS	:
//		Q_BIT		: Quantized bit width per color plane
//		PLANE		: Color plane count
//		SIZE_BIT	: Size setting signal bit width
//
// =============================================================================
// I/O PORTS	:
//		clk					: Clock for all circuit
//		rst_n				: Asynchronous reset ( low active )
//		srst				: Synchronous reset
//
//		valid_i				: Data valid of input Avalon-ST packet
//		startofpacket_i		: Start of packet signal of input Avalon-ST packet
//		endofpacket_i		: End of packet signal of input Avalon-ST packet
//		data_i				: Data of Avalon-ST input Avalon-ST packet
//		ready_o				: Data reception ready of input Avalon-ST packet
//
//		valid_o				: Data valid of output VSI frame data.
//		frame_start_o		: Frame start of output VSI frame data.
//		frame_end_o			: Frame end of output VSI frame data.
//		field_o				: Field status of output VSI frame data.
//							  0:Odd field  1:Even field
//		pixel_o				: Pixel data of output VSI frame data.
//		width_o				: Horizontal size of output VSI frame data
//		height_o			: Vertical size of output frame data
//		ready_i				: Data reception ready of output VSI frame data.
//
//
// =============================================================================
`timescale 1ps/1ps
`default_nettype none

module avb_pkt_asm_avs2vsi (
	clk					,
	rst_n				,
	srst				,

	valid_i				,
	startofpacket_i		,
	endofpacket_i		,
	data_i				,
	ready_o				,

	valid_o				,
	frame_start_o		,
	frame_end_o			,
	field_o				,
	pixel_o				,
	width_o				,
	height_o			,
	ready_i				,

	av_interlacing_o
) ;


// =============================================================================
// PARAMETER DEFINITION
// =============================================================================

	// ---------------------------------------------------------------------
	// Below parameters have to be defined from upper module
	// ---------------------------------------------------------------------
	parameter	Q_BIT				= 8 ;
	parameter	PLANE				= 2 ;
	parameter	SIZE_BIT			= 11 ;

	// ---------------------------------------------------------------------
	// Please do not change the following parameters
	// ---------------------------------------------------------------------
	parameter	PIXEL_BIT			= Q_BIT * PLANE ;
	parameter	BUF_DBIT			= PIXEL_BIT + 2 ;					// vvv Dummy
	parameter	BUF_CBIT			= ( PLANE == 2 ) ?	SIZE_BIT + 16 + 4 + 4 :
														SIZE_BIT + 16 + 4 ;

	parameter	IDLE				= 3'h0 ;
	parameter	GET_SYMBOL			= 3'h1 ;
	parameter	VIDEO_RDY			= 3'h2 ;
	parameter	RDY_WAIT			= 3'h3 ;
	parameter	VIDEO_NRDY			= 3'h4 ;


// =============================================================================
// PORT DECLARATION
// =============================================================================

	input	wire							clk					;
	input	wire							rst_n				;
	input	wire							srst				;

	input	wire							valid_i				;
	input	wire							startofpacket_i		;
	input	wire							endofpacket_i		;
	input	wire	[ PIXEL_BIT-1 : 0 ]		data_i				;
	output	wire							ready_o				;

	output	wire							valid_o				;
	output	wire							frame_start_o		;
	output	wire							frame_end_o			;
	output	wire							field_o				;
	output	wire	[ PIXEL_BIT-1 : 0 ]		pixel_o				;
	output	wire	[  SIZE_BIT-1 : 0 ]		width_o				;
	output	wire	[  SIZE_BIT-1 : 0 ]		height_o			;
	input	wire							ready_i				;

	output	wire	[			3 : 0 ]		av_interlacing_o	;


// =============================================================================
// REG / WIRE DECLARATION
// =============================================================================

	wire							control_packet		;
	wire							video_packet		;
	wire							end_packet			;

	reg		[			2 : 0 ]		av_state_ff			;
	reg								ready_ff			;
	reg								buf_wt_ff			;
	reg		[  BUF_DBIT-1 : 0 ]		buf_wdata_ff		;
	reg		[			3 : 0 ]		symbol_cnt_ff		;
	reg								ready_delay_ff		;

	reg		[  BUF_CBIT-1 : 0 ]		ctl_packet_buf_ff	;
	wire	[  SIZE_BIT-1 : 0 ]		width_buf			;
	wire	[  SIZE_BIT-1 : 0 ]		height_buf			;
	wire	[			3 : 0 ]		field_buf			;
	wire							field				;

	reg		[			2 : 0 ]		wt_point_ff			;
	reg		[			2 : 0 ]		rd_point_ff			;
	wire							buf_full			;
	wire							buf_afull			;
	wire							buf_empty			;

	reg		[  BUF_DBIT-1 : 0 ]		buf_ff				[ 0 : 7 ] ;

	reg								valid_ff			;
	reg								frame_start_ff		;
	reg								frame_end_ff		;
	reg		[ PIXEL_BIT-1 : 0 ]		pixel_ff			;
	reg		[			3 : 0 ]		avfld_status_ff		;
	reg								field_ff			;
	reg		[  SIZE_BIT-1 : 0 ]		width_ff			;
	reg		[  SIZE_BIT-1 : 0 ]		height_ff			;


// =============================================================================
// FUNCTION DESCRIPTION
// =============================================================================

	///////////////////////
	// Avalon-ST control //
	///////////////////////
	assign	control_packet	= ( valid_i & startofpacket_i & ( data_i[3:0]==4'hF ) ) ;
	assign	video_packet	= ( valid_i & startofpacket_i & ( data_i[3:0]==4'h0 ) ) ;
	assign	end_packet		= ( valid_i & endofpacket_i ) ;

	// state machine
	always @( posedge clk or negedge rst_n ) begin
		if ( !rst_n ) begin
			av_state_ff		<= IDLE ;
			ready_ff		<= 1'b0 ;
			buf_wt_ff		<= 1'b0 ;
			buf_wdata_ff	<= { BUF_DBIT{1'b0} } ;
			symbol_cnt_ff	<= 4'h0 ;
		end
		else if ( srst ) begin
			av_state_ff		<= IDLE ;
			ready_ff		<= 1'b0 ;
			buf_wt_ff		<= 1'b0 ;
			buf_wdata_ff	<= { BUF_DBIT{1'b0} } ;
			symbol_cnt_ff	<= 4'h0 ;
		end
		else begin
			case ( av_state_ff )
				IDLE		: begin
					if ( control_packet && ready_delay_ff ) begin
						av_state_ff		<= GET_SYMBOL ;
						ready_ff		<= 1'b1 ;
						buf_wt_ff		<= 1'b0 ;
					end
					else if ( video_packet && ready_delay_ff ) begin
						if ( buf_afull ) begin
							av_state_ff		<= VIDEO_NRDY ;
							ready_ff		<= 1'b0 ;
							buf_wt_ff		<= valid_i ;
							buf_wdata_ff	<= { startofpacket_i , endofpacket_i , data_i } ;
						end
						else begin
							av_state_ff		<= VIDEO_RDY ;
							ready_ff		<= 1'b1 ;
							buf_wt_ff		<= valid_i ;
							buf_wdata_ff	<= { startofpacket_i , endofpacket_i , data_i } ;
						end
					end
					else begin
						ready_ff		<= ~buf_afull ;
						buf_wt_ff		<= 1'b0 ;
					end
				end

				GET_SYMBOL	: begin
					if ( valid_i ) begin
						if ( ( ( PLANE == 1 ) && ( symbol_cnt_ff == 4'h8 ) ) ||
							 ( ( PLANE == 2 ) && ( symbol_cnt_ff == 4'h4 ) ) ||
							 ( ( PLANE == 3 ) && ( symbol_cnt_ff == 4'h2 ) ) ) begin
							av_state_ff		<= IDLE ;
							symbol_cnt_ff	<= 4'h0 ;
							ready_ff		<= ~buf_afull ;
						end
						else begin
							symbol_cnt_ff	<= symbol_cnt_ff + 1'b1 ;
						end
					end
				end

				VIDEO_RDY	: begin
					buf_wdata_ff	<= { startofpacket_i , endofpacket_i , data_i } ;
					if ( control_packet ) begin
						av_state_ff	<= GET_SYMBOL ;
						ready_ff	<= 1'b1 ;
						buf_wt_ff	<= 1'b0 ;
					end
					else if ( end_packet ) begin
						av_state_ff	<= IDLE ;
						ready_ff	<= 1'b1 ;
						buf_wt_ff	<= valid_i & ready_delay_ff ;
					end
					else if ( buf_afull && valid_i ) begin
						av_state_ff	<= RDY_WAIT	;
						ready_ff	<= 1'b0 ;
						buf_wt_ff	<= valid_i & ready_delay_ff ;
					end
					else begin
						ready_ff	<= 1'b1 ;
						buf_wt_ff	<= valid_i & ready_delay_ff ;
					end
				end

				RDY_WAIT	: begin
					buf_wdata_ff	<= { startofpacket_i , endofpacket_i , data_i } ;
					ready_ff		<= ~buf_afull ;
					if ( control_packet ) begin
						av_state_ff	<= GET_SYMBOL ;
						ready_ff	<= 1'b1 ;
						buf_wt_ff	<= 1'b0 ;
					end
					else if ( end_packet ) begin
						av_state_ff	<= IDLE ;
						buf_wt_ff	<= valid_i & ready_delay_ff ;
					end
					else if ( !buf_afull ) begin
						av_state_ff	<= VIDEO_RDY ;
						buf_wt_ff	<= valid_i & ready_delay_ff ;
					end
					else begin
						av_state_ff	<= VIDEO_NRDY ;
						buf_wt_ff	<= valid_i & ready_delay_ff ;
					end
				end

				VIDEO_NRDY	: begin
					buf_wt_ff	<= 1'b0 ;
					if ( !buf_afull ) begin
						av_state_ff	<= VIDEO_RDY ;
						ready_ff	<= 1'b1 ;
					end
				end

				default		: begin
					av_state_ff		<= IDLE ;
					ready_ff		<= 1'b0 ;
					buf_wt_ff		<= 1'b0 ;
					buf_wdata_ff	<= { BUF_DBIT{1'b0} } ;
					symbol_cnt_ff	<= 4'h0 ;
				end
			endcase
		end
	end

	always @( posedge clk or negedge rst_n ) begin
		if ( !rst_n ) begin
			ready_delay_ff	<= 1'b0 ;
		end
		else begin
			ready_delay_ff	<= ready_ff ;
		end
	end

	// control packet latch
	generate
		if ( PLANE == 3 ) begin			: plane3_ctl
			always @( posedge clk or negedge rst_n ) begin
				if ( !rst_n ) begin
					ctl_packet_buf_ff	<= { BUF_CBIT{1'b0} } ;
				end
				else if ( srst ) begin
					ctl_packet_buf_ff	<= { BUF_CBIT{1'b0} } ;
				end
				else begin
					if ( av_state_ff == GET_SYMBOL && valid_i ) begin
						ctl_packet_buf_ff	<= { ctl_packet_buf_ff[BUF_CBIT-12-1:0] , 
											     data_i[3:0] , 
											     data_i[Q_BIT+3:Q_BIT] , 
											     data_i[Q_BIT*2+3:Q_BIT*2] } ;
					end
				end
			end

			assign	width_buf	= ctl_packet_buf_ff[SIZE_BIT+20-1:20] ;
			assign	height_buf	= ctl_packet_buf_ff[SIZE_BIT+4-1:4] ;
			assign	field_buf	= ctl_packet_buf_ff[3:0] ;
			assign	field		= field_buf[2] ;
		end
		else if ( PLANE == 2 ) begin	: plane2_ctl
			always @( posedge clk or negedge rst_n ) begin
				if ( !rst_n ) begin
					ctl_packet_buf_ff	<= { BUF_CBIT{1'b0} } ;
				end
				else if ( srst ) begin
					ctl_packet_buf_ff	<= { BUF_CBIT{1'b0} } ;
				end
				else begin
					if ( av_state_ff == GET_SYMBOL && valid_i ) begin
						ctl_packet_buf_ff	<= { ctl_packet_buf_ff[BUF_CBIT-8-1:0] , data_i[3:0] , data_i[Q_BIT+3:Q_BIT] } ;
					end
				end
			end

			assign	width_buf	= ctl_packet_buf_ff[SIZE_BIT+24-1:24] ;
			assign	height_buf	= ctl_packet_buf_ff[SIZE_BIT+8-1:8] ;
			assign	field_buf	= ctl_packet_buf_ff[7:4] ;
			assign	field		= field_buf[2] ;
		end
		else if ( PLANE == 1 ) begin	: plane1_ctl
			always @( posedge clk or negedge rst_n ) begin
				if ( !rst_n ) begin
					ctl_packet_buf_ff	<= { BUF_CBIT{1'b0} } ;
				end
				else if ( srst ) begin
					ctl_packet_buf_ff	<= { BUF_CBIT{1'b0} } ;
				end
				else begin
					if ( av_state_ff == GET_SYMBOL && valid_i ) begin
						ctl_packet_buf_ff	<= { ctl_packet_buf_ff[BUF_CBIT-4-1:0] , data_i[3:0] } ;
					end
				end
			end

			assign	width_buf	= ctl_packet_buf_ff[SIZE_BIT+20-1:20] ;
			assign	height_buf	= ctl_packet_buf_ff[SIZE_BIT+4-1:4] ;
			assign	field_buf	= ctl_packet_buf_ff[3:0] ;
			assign	field		= field_buf[2] ;
		end
	endgenerate

	////////////
	// buffer //
	////////////
	// write pointer
	always @( posedge clk or negedge rst_n ) begin
		if ( !rst_n ) begin
			wt_point_ff	<= 3'h0 ;
		end
		else if ( srst ) begin
			wt_point_ff	<= 3'h0 ;
		end
		else begin
			if ( buf_wt_ff && !buf_full ) begin
				wt_point_ff	<= wt_point_ff + 1'b1 ;
			end
		end
	end

	// read pointer
	always @( posedge clk or negedge rst_n ) begin
		if ( !rst_n ) begin
			rd_point_ff	<= 3'h0 ;
		end
		else if ( srst ) begin
			rd_point_ff	<= 3'h0 ;
		end
		else begin
			if ( ready_i && !buf_empty ) begin
				rd_point_ff	<= rd_point_ff + 1'b1 ;
			end
		end
	end

	// flag
	assign	buf_full	= ( ( wt_point_ff + 1'b1 ) == rd_point_ff ) ?	1'b1 : 1'b0 ;
	assign	buf_afull	= ( ( ( wt_point_ff + 3'h4 ) == rd_point_ff ) ||
							( ( wt_point_ff + 3'h3 ) == rd_point_ff ) ||
							( ( wt_point_ff + 3'h2 ) == rd_point_ff ) ||
							( ( wt_point_ff + 3'h1 ) == rd_point_ff ) ) ?	1'b1 : 1'b0 ;
	assign	buf_empty	= ( wt_point_ff == rd_point_ff ) ?	1'b1 : 1'b0 ;

	// buffer
	generate
		genvar	i ;
		for ( i = 0 ; i <= 7 ; i = i + 1 ) begin	: data_buf
			always @( posedge clk or negedge rst_n ) begin
				if ( !rst_n ) begin
					buf_ff[i]	<= { BUF_DBIT{1'b0} } ;
				end
				else if ( srst ) begin
					buf_ff[i]	<= { BUF_DBIT{1'b0} } ;
				end
				else begin
					if ( buf_wt_ff && !buf_full && ( wt_point_ff == i[2:0] ) ) begin
						buf_ff[i]	<= buf_wdata_ff ;
					end
				end
			end
		end
	endgenerate

	/////////////////
	// VSI control //
	/////////////////
	always @( posedge clk or negedge rst_n ) begin
		if ( !rst_n ) begin
			valid_ff		<= 1'b0 ;
			frame_start_ff	<= 1'b0 ;
			frame_end_ff	<= 1'b0 ;
			pixel_ff		<= { PIXEL_BIT{1'b0} } ;
		end
		else if ( srst ) begin
			valid_ff		<= 1'b0 ;
			frame_start_ff	<= 1'b0 ;
			frame_end_ff	<= 1'b0 ;
			pixel_ff		<= { PIXEL_BIT{1'b0} } ;
		end
		else begin
			case ( { ready_i , buf_empty } )
				2'b11	: begin
					valid_ff		<= 1'b0 ;
					frame_start_ff	<= 1'b0 ;
					frame_end_ff	<= 1'b0 ;
					pixel_ff		<= { PIXEL_BIT{1'b0} } ;
				end

				2'b10	: begin
					valid_ff		<= 1'b1 ;
					frame_start_ff	<= buf_ff[rd_point_ff][PIXEL_BIT+1] ;
					frame_end_ff	<= buf_ff[rd_point_ff][PIXEL_BIT  ] ;
					pixel_ff		<= buf_ff[rd_point_ff][PIXEL_BIT-1:0] ;
				end

				default	: begin
					valid_ff		<= valid_ff ;
					frame_start_ff	<= frame_start_ff ;
					frame_end_ff	<= frame_end_ff ;
					pixel_ff		<= pixel_ff ;
				end
			endcase
		end
	end

	always @( posedge clk or negedge rst_n ) begin
		if ( !rst_n ) begin
			avfld_status_ff	<= 4'h0 ;
			field_ff		<= 1'b0 ;
			width_ff		<= { SIZE_BIT{1'b0} } ;
			height_ff		<= { SIZE_BIT{1'b0} } ;
		end
		else if ( srst ) begin
			avfld_status_ff	<= 4'h0 ;
			field_ff		<= 1'b0 ;
			width_ff		<= { SIZE_BIT{1'b0} } ;
			height_ff		<= { SIZE_BIT{1'b0} } ;
		end
		else begin
			if ( ready_i && !buf_empty && buf_ff[rd_point_ff][PIXEL_BIT+1] ) begin
				avfld_status_ff	<= field_buf ;
				field_ff		<= field ;
				width_ff		<= width_buf ;
				height_ff		<= height_buf ;
			end
		end
	end

	// Output signal
	assign	ready_o				= ready_ff ;

	assign	valid_o				= valid_ff ;
	assign	frame_start_o		= frame_start_ff ;
	assign	frame_end_o			= frame_end_ff ;
	assign	field_o				= field_ff ;
	assign	pixel_o				= pixel_ff ;
	assign	width_o				= width_ff ;
	assign	height_o			= height_ff ;
	assign	av_interlacing_o	= avfld_status_ff ;


endmodule
`default_nettype wire