// =============================================================================
// Copyright (C) MACNICA inc.
// =============================================================================
// FILE NAME	: avb_pkt_asm_csr.v
// DEPARTMENT	: MACNICA Strategic Technology Group
// AUTHOR		: Ryosuke Yamamoto
// DATE			: 03/07/2014
// =============================================================================
// DESCRIPTION	:
//
// =============================================================================
// REVISION HISTORY	:
//	v1.0	: Initial version release
//
// =============================================================================
// PARAMETERS	:
//		DVEND 	: dividend bit width 2 ~ 32
//		DVSOR 	: divisor bit width 2 ~ 32
//
// =============================================================================
// I/O PORTS	:
//		clk			: clock for all circuit
//		rst_n		: asynchronous reset ( low active )
//		srst		: synchronous reset
//		enable	 	: clock enable
//		
//		valid_i		: data valid of input data
//		dividend_i	: dividend data
//		divisor_i	: divisor data
//		
//		valid_o		: data valid of output data
//		dividend_o	: shifted dividend data
//		divisor_o	: divisor data
//
// =============================================================================
`timescale 1ps/1ps
`default_nettype none

module	avb_pkt_asm_csr (
	clk					,
	rst_n				,

	address_i			,
	byteenable_i		,
	read_i				,
	write_i				,
	readdatavalid_o		,
	readdata_o			,
	writedata_i			,
	waitrequest_o		,

	packet_generate_i	,

	srst_o				,
	packet_size_o		,
	des_mac_adr_o		,
	src_mac_adr_o		,
	tag_o				,
	ether_type_o		,
	packet_transmit_o	
) ;
	
// =============================================================================
// PARAMETER DEFINITION
// =============================================================================

	// ---------------------------------------------------------------------
	// Below parameters have to be defined from upper module
	// ---------------------------------------------------------------------

	// ---------------------------------------------------------------------
	// Please do not change the following parameters
	// ---------------------------------------------------------------------
	parameter	PKT_SIZE_BIT	= 11 ;
	parameter	MAC_ADR_BIT		= 48 ;
	parameter	TAG_BIT			= 32 ;
	parameter	TYPE_BIT		= 16 ;
	
	parameter	REG_NUM			=  9 ;
	parameter	REG_INI			= { 32'h0000_0000 ,		// address 0x8 initial value
								    32'h0000_0000 ,		// address 0x7 initial value
								    32'h0000_0000 ,		// address 0x6 initial value
								    32'h0000_0000 ,		// address 0x5 initial value
								    32'h0000_0000 ,		// address 0x4 initial value
								    32'h0000_0000 ,		// address 0x3 initial value
								    32'h0000_0000 ,		// address 0x2 initial value
								    32'h0000_0400 ,		// address 0x1 initial value
								    32'h0000_0001 } ;	// address 0x0 initial value
	parameter	REG_VLD			= { 32'h0000_0001 ,		// address 0x8 register valid ( packet_transmit )
								    32'h0000_FFFF ,		// address 0x7 register valid ( ether_type )
								    32'hFFFF_FFFF ,		// address 0x6 register valid ( tag )
								    32'hFFFF_FFFF ,		// address 0x5 register valid ( src_mac_adr L )
								    32'h0000_FFFF ,		// address 0x4 register valid ( src_mac_adr H )
								    32'hFFFF_FFFF ,		// address 0x3 register valid ( des_mac_adr L )
								    32'h0000_FFFF ,		// address 0x2 register valid ( des_mac_adr H )
								    32'h0000_07FF ,		// address 0x1 register valid ( packet_size )
								    32'h0000_0001 } ;	// address 0x0 register valid ( srst )

// =============================================================================
// PORT DECLARATION
// =============================================================================

	input	wire							clk					;
	input	wire							rst_n				;

	input	wire	[              3 : 0 ]	address_i			;
	input	wire	[              3 : 0 ]	byteenable_i		;
	input	wire							read_i				;
	input	wire							write_i				;
	output	wire							readdatavalid_o		;
	output	wire	[             31 : 0 ]	readdata_o			;
	input	wire	[             31 : 0 ]	writedata_i			;
	output	wire							waitrequest_o		;

	input	wire							packet_generate_i	;

	output	wire							srst_o				;
	output	wire	[ PKT_SIZE_BIT-1 : 0 ]	packet_size_o		;
	output	wire	[  MAC_ADR_BIT-1 : 0 ]	des_mac_adr_o		;
	output	wire	[  MAC_ADR_BIT-1 : 0 ]	src_mac_adr_o		;
	output	wire	[      TAG_BIT-1 : 0 ]	tag_o				;
	output	wire	[     TYPE_BIT-1 : 0 ]	ether_type_o		;
	output	wire							packet_transmit_o	;
	
// =============================================================================
// REG / WIRE DECLARATION
// =============================================================================
	
	genvar				i					;

	reg		[ 31 : 0 ]	reg_ff				[ 0 : REG_NUM-1 ] ;
	reg		[ 31 : 0 ]	rd_data_ff			;
	reg					waitrequest_ff		;
	reg					readdatavalid_ff	;
	
// =============================================================================
// FUNCTION DESCRIPTION
// =============================================================================
	
	generate
		for ( i = 0 ; i < REG_NUM ; i = i + 1 ) begin		: generate_register
			always @ ( posedge clk or negedge rst_n ) begin
				if ( !rst_n ) begin
					reg_ff[i]	<= REG_INI[i*32+31:i*32] ;
				end
				else begin
					if ( write_i && address_i == i[3:0] && !waitrequest_ff ) begin
						reg_ff[i][ 7: 0]	<= ( byteenable_i[0] ) ?	writedata_i[ 7: 0] & REG_VLD[i*32+ 7:i*32   ]	: reg_ff[i][ 7: 0] ;
						reg_ff[i][15: 8]	<= ( byteenable_i[1] ) ?	writedata_i[15: 8] & REG_VLD[i*32+15:i*32+ 8]	: reg_ff[i][15: 8] ;
						reg_ff[i][23:16]	<= ( byteenable_i[2] ) ?	writedata_i[23:16] & REG_VLD[i*32+23:i*32+16]	: reg_ff[i][23:16] ;
						reg_ff[i][31:24]	<= ( byteenable_i[3] ) ?	writedata_i[31:24] & REG_VLD[i*32+31:i*32+24]	: reg_ff[i][31:24] ;
					end
				end
			end
		end
	endgenerate

	always @ ( posedge clk or negedge rst_n ) begin
		if ( !rst_n ) begin
			rd_data_ff	<= 32'h0000_0000 ;
		end
		else begin
			if ( address_i == REG_NUM[3:0] ) begin
				rd_data_ff	<= { 31'h0000_0000 , packet_generate_i } ;
			end
			else begin
				rd_data_ff	<= reg_ff[address_i] ;
			end
		end
	end

	always @ ( posedge clk or negedge rst_n ) begin
		if ( !rst_n ) begin
			waitrequest_ff	<= 1'b1 ;
		end
		else begin
			if ( !waitrequest_ff ) begin
				waitrequest_ff	<= 1'b1 ;
			end
			else if ( read_i || write_i ) begin
				waitrequest_ff	<= 1'b0 ;
			end
		end
	end

	always @ ( posedge clk or negedge rst_n ) begin
		if ( !rst_n ) begin
			readdatavalid_ff	<= 1'b0 ;
		end
		else begin
			readdatavalid_ff	<= ~waitrequest_ff & read_i ;
		end
	end

	assign	readdatavalid_o		= readdatavalid_ff ;
	assign	readdata_o			= rd_data_ff ;
	assign	waitrequest_o		= waitrequest_ff ;

	assign	srst_o				= reg_ff[4'h0][0] ;
	assign	packet_size_o		= reg_ff[4'h1][PKT_SIZE_BIT-1:0] ;
	assign	des_mac_adr_o		= { reg_ff[4'h2][15:0] , reg_ff[4'h3] } ;
	assign	src_mac_adr_o		= { reg_ff[4'h4][15:0] , reg_ff[4'h5] } ;
	assign	tag_o				= reg_ff[4'h6][31:0] ;
	assign	ether_type_o		= reg_ff[4'h7][15:0] ;
	assign	packet_transmit_o	= reg_ff[4'h8][0] ;


endmodule

`default_nettype wire
