// =============================================================================
// Copyright (C) MACNICA inc.
// =============================================================================
// FILE NAME	: avb_pkt_asm_div.v
// DEPARTMENT	: MACNICA Strategic Technology Group
// AUTHOR		: Ryosuke Yamamoto
// DATE 		: 03/05/2014
// =============================================================================
// DESCRIPTION	:
//		
// =============================================================================
// REVISION HISTORY :
//		v1.0		: Initial Release
//
// =============================================================================
// PARAMETERS	:
//		Q_BIT		: Quantized bit width per color plane
//		PLANE		: Color plane count
//		SIZE_BIT	: Size setting signal bit width
//
// =============================================================================
// I/O PORTS	:
//		clk						: Clock for all circuit
//		rst_n					: Asynchronous reset ( low active )
//		srst					: Synchronous reset
//
//		valid_i					: Data valid of input VSI frame data.
//		frame_start_i			: Frame start of input VSI frame data.
//		frame_end_i				: Frame end of input VSI frame data.
//		field_i					: Field status of input VSI frame data.
//								  0:Odd field  1:Even field
//		data_i					: Pixel data of input VSI frame data.
//		width_i					: Horizontal size of input VSI frame data
//		height_i				: Vertical size of input frame data
//		ready_o					: Data reception ready of input VSI frame data.
//
//		valid_o					: Data valid of output VSI frame data.
//		frame_start_o			: Frame start of output VSI frame data.
//		frame_end_o				: Frame end of output VSI frame data.
//		field_o					: Field status of output VSI frame data.
//								  0:Odd field  1:Even field
//		data_o					: Pixel data of output VSI frame data.
//		width_o					: Horizontal size of output VSI frame data
//		height_o				: Vertical size of output frame data
//		packet_number_o			: Packet Number of output frame data
//		total_packet_number_o	: Total packet number of output frame data
//		ready_i					: Data reception ready of output VSI frame data.
//
//
// =============================================================================
`timescale 1ps/1ps
`default_nettype none

module avb_pkt_asm_div (
	clk						,
	rst_n					,
	srst					,

	packet_size_i			,

	valid_i					,
	frame_start_i			,
	frame_end_i				,
	field_i					,
	data_i					,
	width_i					,
	height_i				,
	ready_o					,

	valid_o					,
	frame_start_o			,
	frame_end_o				,
	field_o					,
	data_o					,
	width_o					,
	height_o				,
	frame_num_o				,
	packet_number_o			,
	total_packet_count_o	,
	ready_i				

) ;


// =============================================================================
// PARAMETER DEFINITION
// =============================================================================

	// ---------------------------------------------------------------------
	// Below parameters have to be defined from upper module
	// ---------------------------------------------------------------------
	parameter	Q_BIT				= 8 ;
	parameter	PLANE				= 2 ;
	parameter	SIZE_BIT			= 11 ;

	// ---------------------------------------------------------------------
	// Please do not change the following parameters
	// ---------------------------------------------------------------------
	parameter	IN_DATA_BIT			= Q_BIT * PLANE ;
	parameter	DATA_BIT			= 32 ;
	parameter	PKT_SIZE_BIT		= 11 ;
	parameter	T_PKT_SIZE_BIT		= SIZE_BIT * 2 ;
	parameter	FRM_NUM_BIT			= 16 ;
	parameter	PACKET_NUM_BIT		= 32 ;

// =============================================================================
// PORT DECLARATION
// =============================================================================

	input	wire								clk						;
	input	wire								rst_n					;
	input	wire								srst					;

	input	wire	[   PKT_SIZE_BIT-1 : 0 ]	packet_size_i			;	// A multiple of 4

	input	wire								valid_i					;
	input	wire								frame_start_i			;
	input	wire								frame_end_i				;
	input	wire								field_i					;
	input	wire	[       DATA_BIT-1 : 0 ]	data_i					;
	input	wire	[       SIZE_BIT-1 : 0 ]	width_i					;
	input	wire	[       SIZE_BIT-1 : 0 ]	height_i				;
	output	wire								ready_o					;

	output	wire								valid_o					;
	output	wire								frame_start_o			;
	output	wire								frame_end_o				;
	output	wire								field_o					;
	output	wire	[       DATA_BIT-1 : 0 ]	data_o					;
	output	wire	[       SIZE_BIT-1 : 0 ]	width_o					;
	output	wire	[       SIZE_BIT-1 : 0 ]	height_o				;
	output	wire	[    FRM_NUM_BIT-1 : 0 ]	frame_num_o				;
	output	wire	[ PACKET_NUM_BIT-1 : 0 ]	packet_number_o			;
	output	wire	[ PACKET_NUM_BIT-1 : 0 ]	total_packet_count_o	;
	input	wire								ready_i					;


// =============================================================================
// REG / WIRE DECLARATION
// =============================================================================

	wire								frame_start				;
	wire								frame_end				;

	reg									ready_ff				;
	reg		[    FRM_NUM_BIT-1 : 0 ]	frame_num_ff			;
	reg									div_calc_ff				;
	reg									div_calc_interval_ff	;
	reg		[     SIZE_BIT*2-1 : 0 ]	dividend_ff				;
	reg		[   PKT_SIZE_BIT-1 : 0 ]	divisor_ff				;
	reg		[   PKT_SIZE_BIT-1 : 0 ]	packet_cnt_max_ff		;
	reg		[   T_PKT_SIZE_BIT : 0 ]	total_packet_count_ff	;
	wire								div_result				;
	wire	[   T_PKT_SIZE_BIT : 0 ]	div_quotient			;
	wire	[       SIZE_BIT-1 : 0 ]	div_remainder			;
//	wire	[       SIZE_BIT*2 : 0 ]	div_remainder			;
	reg									valid_frame_ff			;
	reg		[   PKT_SIZE_BIT-1 : 0 ]	data_cnt_ff				;
	wire								data_cnt_max			;
	reg									valid_pre_ff			;
	reg									frame_start_pre_ff		;
	reg									frame_end_pre_ff		;
	reg		[       DATA_BIT-1 : 0 ]	data_pre_ff				;
	reg									field_pre_ff			;
	reg		[       SIZE_BIT-1 : 0 ]	width_pre_ff			;
	reg		[       SIZE_BIT-1 : 0 ]	height_pre_ff			;
	reg		[ T_PKT_SIZE_BIT-1 : 0 ]	packet_number_ff		;
	reg									valid_lat_ff			;
	reg									frame_start_lat_ff		;
	reg									frame_end_lat_ff		;
	reg		[       DATA_BIT-1 : 0 ]	data_lat_ff				;
	reg									field_lat_ff			;
	reg		[       SIZE_BIT-1 : 0 ]	width_lat_ff			;
	reg		[       SIZE_BIT-1 : 0 ]	height_lat_ff			;
	reg									valid_ff				;
	reg									frame_start_ff			;
	reg									frame_end_ff			;
	reg		[       DATA_BIT-1 : 0 ]	data_ff					;
	reg									field_ff				;
	reg		[       SIZE_BIT-1 : 0 ]	width_ff				;
	reg		[       SIZE_BIT-1 : 0 ]	height_ff				;


// =============================================================================
// FUNCTION DESCRIPTION
// =============================================================================

	assign	frame_start	= valid_i & frame_start_i & ~( |data_i[3:0] ) ;
	assign	frame_end	= valid_i & frame_end_i ;

	always @ ( posedge clk or negedge rst_n ) begin
		if ( !rst_n ) begin
			ready_ff	<= 1'b0 ;
		end
		else if ( srst ) begin
			ready_ff	<= 1'b0 ;
		end
		else begin
			if ( ready_ff && frame_start ) begin
				ready_ff	<= 1'b0 ;
			end
			else if ( !div_calc_interval_ff ) begin
				ready_ff	<= ready_i ;
			end
		end
	end

	always @ ( posedge clk or negedge rst_n ) begin
		if ( !rst_n ) begin
			frame_num_ff	<= { FRM_NUM_BIT{1'b0} } ;
		end
		else if ( srst ) begin
			frame_num_ff	<= { FRM_NUM_BIT{1'b0} } ;
		end
		else begin
			if ( ready_ff && frame_start ) begin
				frame_num_ff	<= frame_num_ff + 1'b1 ;
			end
		end
	end

	//------------------------------------//
	// division process for packet number //
	//------------------------------------//
	always @ ( posedge clk or negedge rst_n ) begin
		if ( !rst_n ) begin
			div_calc_ff				<= 1'b0 ;
			div_calc_interval_ff	<= 1'b0 ;
		end
		else if ( srst ) begin
			div_calc_ff				<= 1'b0 ;
			div_calc_interval_ff	<= 1'b0 ;
		end
		else begin
			div_calc_ff	<= ready_ff & frame_start ;
			if ( ready_ff && frame_start ) begin
				div_calc_interval_ff	<= 1'b1 ;
			end
			else if ( div_result ) begin
				div_calc_interval_ff	<= 1'b0 ;
			end
		end
	end

	always @ ( posedge clk or negedge rst_n ) begin
		if ( !rst_n ) begin
			dividend_ff			<= { (SIZE_BIT*2){1'b0} } ;
			divisor_ff			<= { PKT_SIZE_BIT{1'b0} } ;
			packet_cnt_max_ff	<= { PKT_SIZE_BIT{1'b1} } ;
		end
		else if ( srst ) begin
			dividend_ff			<= { (SIZE_BIT*2){1'b0} } ;
			divisor_ff			<= { PKT_SIZE_BIT{1'b0} } ;
			packet_cnt_max_ff	<= { PKT_SIZE_BIT{1'b1} } ;
		end
		else begin
			if ( ready_ff && frame_start ) begin
				dividend_ff			<= width_i * height_i ;
				divisor_ff			<= packet_size_i ;
				packet_cnt_max_ff	<= { 2'b00 , packet_size_i[PKT_SIZE_BIT-1:2] } - 1'b1 ;
			end
		end
	end

	always @ ( posedge clk or negedge rst_n ) begin
		if ( !rst_n ) begin
			total_packet_count_ff	<= { T_PKT_SIZE_BIT+1{1'b0} } ;
		end
		else if ( srst ) begin
			total_packet_count_ff	<= { T_PKT_SIZE_BIT+1{1'b0} } ;
		end
		else begin
			if ( div_result ) begin
				total_packet_count_ff	<= div_quotient + |div_remainder ;
			end
		end
	end

	// divider module 
	avb_pkt_asm_divider
		#(
			.DVEND			( ( SIZE_BIT * 2 ) + 1	) ,
			.DVSOR			( PKT_SIZE_BIT			) ,
			.PIPELINE		( 0						) 
		)
		u_divider (
			.clk			( clk						) ,
			.rst_n			( rst_n						) ,
			.srst			( 1'b0						) ,
			.enable			( 1'b1						) ,
			
			.valid_i		( div_calc_ff				) ,
			.dividend_i		( { dividend_ff , 1'b0 }	) ,
			.divisor_i		( divisor_ff				) ,
			
			.ready_o		( 							) ,
			.valid_o		( div_result				) ,
			.quotient_o		( div_quotient				) ,
			.remainder_o	( div_remainder				) 
		) ;

	//----------------//
	// frame Division //
	//----------------//
	always @ ( posedge clk or negedge rst_n ) begin
		if ( !rst_n ) begin
			valid_frame_ff	<= 1'b0 ;
		end
		else if ( srst ) begin
			valid_frame_ff	<= 1'b0 ;
		end
		else begin
			if ( ready_ff ) begin
				if ( frame_start ) begin
					valid_frame_ff	<= 1'b1 ;
				end
				else if ( frame_end ) begin
					valid_frame_ff	<= 1'b0 ;
				end
			end
		end
	end

	always @ ( posedge clk or negedge rst_n ) begin
		if ( !rst_n ) begin
			data_cnt_ff	<= { PKT_SIZE_BIT{1'b0} } ;
		end
		else if ( srst ) begin
			data_cnt_ff	<= { PKT_SIZE_BIT{1'b0} } ;
		end
		else begin
			if ( ready_ff ) begin
				if ( frame_start || ( valid_frame_ff && frame_end_pre_ff ) ) begin
					data_cnt_ff	<= { PKT_SIZE_BIT{1'b0} } ;
				end
				else if ( valid_i ) begin
					data_cnt_ff	<= data_cnt_ff + 1'b1 ;
				end
			end
		end
	end

	assign	data_cnt_max	= ( data_cnt_ff == packet_cnt_max_ff ) ?	1'b1	 : 1'b0 ;

	always @ ( posedge clk or negedge rst_n ) begin
		if ( !rst_n ) begin
			frame_start_pre_ff	<= 1'b0 ;
			valid_pre_ff		<= 1'b0 ;
		end
		else if ( srst ) begin
			frame_start_pre_ff	<= 1'b0 ;
			valid_pre_ff		<= 1'b0 ;
		end
		else begin
			if ( ready_ff ) begin
				if ( frame_start || ( valid_frame_ff && frame_end_pre_ff ) ) begin
					frame_start_pre_ff	<= 1'b1 ;
					valid_pre_ff		<= 1'b1 ;
				end
				else begin
					frame_start_pre_ff	<= 1'b0 ;
					valid_pre_ff		<= valid_i ;
				end
			end
		end
	end

	always @ ( posedge clk or negedge rst_n ) begin
		if ( !rst_n ) begin
			frame_end_pre_ff	<= 1'b0 ;
		end
		else if ( srst ) begin
			frame_end_pre_ff	<= 1'b0 ;
		end
		else begin
			if ( ready_ff ) begin
				if ( frame_end || ( data_cnt_max && valid_i ) )begin
					frame_end_pre_ff	<= 1'b1 ;
				end
				else begin
					frame_end_pre_ff	<= 1'b0 ;
				end
			end
		end
	end

	always @ ( posedge clk or negedge rst_n ) begin
		if ( !rst_n ) begin
			data_pre_ff	<= { DATA_BIT{1'b0} } ;
		end
		else if ( srst ) begin
			data_pre_ff	<= { DATA_BIT{1'b0} } ;
		end
		else begin
			if ( ready_ff ) begin
				if ( frame_start || ( valid_frame_ff && frame_end_pre_ff ) ) begin
					data_pre_ff	<= { DATA_BIT{1'b0} } ;
				end
				else if ( valid_i ) begin
					data_pre_ff	<= data_i ;
				end
			end
		end
	end

	always @ ( posedge clk or negedge rst_n ) begin
		if ( !rst_n ) begin
			field_pre_ff		<= 1'b0 ;
			width_pre_ff		<= { SIZE_BIT{1'b0} } ;
			height_pre_ff		<= { SIZE_BIT{1'b0} } ;
			packet_number_ff	<= { { T_PKT_SIZE_BIT-1{1'b0} } , 1'b1 } ;
		end
		else if ( srst ) begin
			field_pre_ff		<= 1'b0 ;
			width_pre_ff		<= { SIZE_BIT{1'b0} } ;
			height_pre_ff		<= { SIZE_BIT{1'b0} } ;
			packet_number_ff	<= { { T_PKT_SIZE_BIT-1{1'b0} } , 1'b1 } ;
		end
		else begin
			if ( ready_ff ) begin
				if ( frame_start ) begin
					field_pre_ff		<= 1'b0 ;
					width_pre_ff		<= { SIZE_BIT{1'b0} } ;
					height_pre_ff		<= { SIZE_BIT{1'b0} } ;
					packet_number_ff	<= { { T_PKT_SIZE_BIT-1{1'b0} } , 1'b1 } ;
				end
				else if ( valid_frame_ff && frame_end_pre_ff ) begin
					packet_number_ff	<= packet_number_ff + 1'b1 ;
				end
			end
		end
	end

	always @ ( posedge clk or negedge rst_n ) begin
		if ( !rst_n ) begin
			valid_lat_ff			<= 1'b0 ;
			frame_start_lat_ff		<= 1'b0 ;
			frame_end_lat_ff		<= 1'b0 ;
			data_lat_ff				<= { DATA_BIT{1'b0} } ;
			field_lat_ff			<= 1'b0 ;
			width_lat_ff			<= { SIZE_BIT{1'b0} } ;
			height_lat_ff			<= { SIZE_BIT{1'b0} } ;
		end
		else if ( srst ) begin
			valid_lat_ff			<= 1'b0 ;
			frame_start_lat_ff		<= 1'b0 ;
			frame_end_lat_ff		<= 1'b0 ;
			data_lat_ff				<= { DATA_BIT{1'b0} } ;
			field_lat_ff			<= 1'b0 ;
			width_lat_ff			<= { SIZE_BIT{1'b0} } ;
			height_lat_ff			<= { SIZE_BIT{1'b0} } ;
		end
		else begin
			if ( ready_ff ) begin
				valid_lat_ff			<= valid_pre_ff ;
				frame_start_lat_ff		<= frame_start_pre_ff ;
				frame_end_lat_ff		<= frame_end_pre_ff ;
				data_lat_ff				<= data_pre_ff ;
				field_lat_ff			<= field_pre_ff ;
				width_lat_ff			<= width_pre_ff ;
				height_lat_ff			<= height_pre_ff ;
			end
		end
	end

	always @ ( posedge clk or negedge rst_n ) begin
		if ( !rst_n ) begin
			valid_ff			<= 1'b0 ;
			frame_start_ff		<= 1'b0 ;
			frame_end_ff		<= 1'b0 ;
			data_ff				<= { DATA_BIT{1'b0} } ;
			field_ff			<= 1'b0 ;
			width_ff			<= { SIZE_BIT{1'b0} } ;
			height_ff			<= { SIZE_BIT{1'b0} } ;
		end
		else if ( srst ) begin
			valid_ff			<= 1'b0 ;
			frame_start_ff		<= 1'b0 ;
			frame_end_ff		<= 1'b0 ;
			data_ff				<= { DATA_BIT{1'b0} } ;
			field_ff			<= 1'b0 ;
			width_ff			<= { SIZE_BIT{1'b0} } ;
			height_ff			<= { SIZE_BIT{1'b0} } ;
		end
		else begin
			if ( ready_i ) begin
				valid_ff			<= ( ready_ff ) ?	valid_pre_ff		: valid_lat_ff ;
				frame_start_ff		<= ( ready_ff ) ?	frame_start_pre_ff	: frame_start_lat_ff ;
				frame_end_ff		<= ( ready_ff ) ?	frame_end_pre_ff	: frame_end_lat_ff ;
				data_ff				<= ( ready_ff ) ?	data_pre_ff			: data_lat_ff ;
				field_ff			<= ( ready_ff ) ?	field_pre_ff		: field_lat_ff ;
				width_ff			<= ( ready_ff ) ?	width_pre_ff		: width_lat_ff ;
				height_ff			<= ( ready_ff ) ?	height_pre_ff		: height_lat_ff ;
			end
		end
	end


	// Output signal
	assign	ready_o					= ready_ff ;

	assign	valid_o					= valid_ff ;
	assign	frame_start_o			= frame_start_ff ;
	assign	frame_end_o				= frame_end_ff ;
	assign	field_o					= field_ff ;
	assign	data_o					= data_ff ;
	assign	width_o					= width_ff ;
	assign	height_o				= height_ff ;
	assign	frame_num_o				= frame_num_ff ;
	assign	packet_number_o			= { { (PACKET_NUM_BIT-T_PKT_SIZE_BIT){1'b0} } , packet_number_ff } ;
	assign	total_packet_count_o	= { { (PACKET_NUM_BIT-T_PKT_SIZE_BIT){1'b0} } , total_packet_count_ff } ;


endmodule
`default_nettype wire