// =============================================================================
// Copyright (C) MACNICA inc.
// =============================================================================
// FILE NAME	: avb_pkt_asm_packet.v
// DEPARTMENT	: MACNICA Strategic Technology Group
// AUTHOR		: Ryosuke Yamamoto
// DATE 		: 03/05/2014
// =============================================================================
// DESCRIPTION	:
//		
// =============================================================================
// REVISION HISTORY :
//		v1.0		: Initial Release
//
// =============================================================================
// PARAMETERS	:
//		Q_BIT		: Quantized bit width per color plane
//		PLANE		: Color plane count
//		SIZE_BIT	: Size setting signal bit width
//
// =============================================================================
// I/O PORTS	:
//		clk					: Clock for all circuit
//		rst_n				: Asynchronous reset ( low active )
//		srst				: Synchronous reset
//
//
//
// =============================================================================
`timescale 1ps/1ps
`default_nettype none

module avb_pkt_asm_packet (
	clk							,
	rst_n						,
	srst						,

	des_mac_adr_i				,
	src_mac_adr_i				,
	tag_i						,
	ether_type_i				,

	local_sync_256ns_pulse_i	, //   (b0) local clock toggles every 256 synchronized ns.  (b1) if as synced
	master_clock_sec_i			, // master time used to gate start of audio playback
	master_clock_ns_i			, //

	packet_transmit_i			,
	packet_generate_o			,

	valid_i						,
	frame_start_i				,
	frame_end_i					,
	field_i						,
	data_i						,
	width_i						,
	height_i					,
	frame_num_i					,
	packet_number_i				,
	total_packet_count_i		,
	ready_o						,

	tx_valid_o					,
	tx_startofpacket_o			,
	tx_endofpacket_o			,
	tx_data_o					,
	tx_mod_o					,
	tx_err_o					,
	tx_ready_i					

) ;


// =============================================================================
// PARAMETER DEFINITION
// =============================================================================

	// ---------------------------------------------------------------------
	// Below parameters have to be defined from upper module
	// ---------------------------------------------------------------------
	parameter	SIZE_BIT		= 11 ;
	// ---------------------------------------------------------------------
	// Please do not change the following parameters
	// ---------------------------------------------------------------------
	parameter	DATA_BIT		= 32 ;
	parameter	MAC_ADR_BIT		= 48 ;
	parameter	TAG_BIT			= 32 ;
	parameter	TYPE_BIT		= 16 ;
	parameter	FRM_NUM_BIT		= 16 ;
	parameter	PACKET_NUM_BIT	= 32 ;
	parameter	T_STP_SEC_BIT	= 48 ;
	parameter	T_STP_NS_BIT	= 32 ;

	parameter	ST_IDLE			= 3'h0 ;
	parameter	ST_STANDBY		= 3'h1 ;
	parameter	ST_ETH_HEAD		= 3'h2 ;
	parameter	ST_PAY_HEAD		= 3'h3 ;
	parameter	ST_PAY_DATA		= 3'h4 ;

// =============================================================================
// PORT DECLARATION
// =============================================================================

	input	wire								clk							;
	input	wire								rst_n						;
	input	wire								srst						;

	input	wire	[    MAC_ADR_BIT-1 : 0 ]	des_mac_adr_i				;
	input	wire	[    MAC_ADR_BIT-1 : 0 ]	src_mac_adr_i				;
	input	wire	[        TAG_BIT-1 : 0 ]	tag_i						;
	input	wire	[       TYPE_BIT-1 : 0 ]	ether_type_i				;

	input	wire	[                1 : 0 ]	local_sync_256ns_pulse_i	; //   (b0) local clock toggles every 256 synchronized ns.  (b1) if as synced
	input	wire	[  T_STP_SEC_BIT-1 : 0 ]	master_clock_sec_i			; // master time used to gate start of audio playback
	input	wire	[   T_STP_NS_BIT-1 : 0 ]	master_clock_ns_i			; //

	input	wire								packet_transmit_i			;
	output	wire								packet_generate_o			;

	input	wire								valid_i						;
	input	wire								frame_start_i				;
	input	wire								frame_end_i					;
	input	wire								field_i						;
	input	wire	[       DATA_BIT-1 : 0 ]	data_i						;
	input	wire	[       SIZE_BIT-1 : 0 ]	width_i						;
	input	wire	[       SIZE_BIT-1 : 0 ]	height_i					;
	input	wire	[    FRM_NUM_BIT-1 : 0 ]	frame_num_i					;
	input	wire	[ PACKET_NUM_BIT-1 : 0 ]	packet_number_i				;
	input	wire	[ PACKET_NUM_BIT-1 : 0 ]	total_packet_count_i		;
	output	wire								ready_o						;

	output	wire								tx_valid_o					;
	output	wire								tx_startofpacket_o			;
	output	wire								tx_endofpacket_o			;
	output	wire	[       DATA_BIT-1 : 0 ]	tx_data_o					;
	output	wire	[                1 : 0 ]	tx_mod_o					;
	output	wire								tx_err_o					;
	input	wire								tx_ready_i					;


// =============================================================================
// REG / WIRE DECLARATION
// =============================================================================

	wire								frame_start			;
	wire								frame_end			;

	reg									ready_ff			;

	reg		[                2 : 0 ]	pkt_gen_state_ff	;
	reg									tx_valid_pre_ff		;
	reg									tx_sop_pre_ff		;
	reg									tx_eop_pre_ff		;
	reg		[       DATA_BIT-1 : 0 ]	tx_data_pre_ff		;
	reg		[    MAC_ADR_BIT-1 : 0 ]	des_mac_adr_ff		;
	reg		[    MAC_ADR_BIT-1 : 0 ]	src_mac_adr_ff		;
	reg		[        TAG_BIT-1 : 0 ]	tag_ff				;
	reg		[       TYPE_BIT-1 : 0 ]	ether_type_ff		;
	reg		[    FRM_NUM_BIT-1 : 0 ]	frame_num_ff		;
	reg		[ PACKET_NUM_BIT-1 : 0 ]	pkt_number_ff		;
	reg		[ PACKET_NUM_BIT-1 : 0 ]	ttl_pkt_count_ff	;
	reg		[  T_STP_SEC_BIT-1 : 0 ]	master_clock_sec_ff	;
	reg		[   T_STP_NS_BIT-1 : 0 ]	master_clock_ns_ff	;
	reg		[                3 : 0 ]	pkt_cnt_ff			;
	reg									packet_generate_ff	;

	reg									ready_dly_ff		;
	
	reg									tx_valid_lat_ff		;
	reg									tx_sop_lat_ff		;
	reg									tx_eop_lat_ff		;
	reg		[       DATA_BIT-1 : 0 ]	tx_data_lat_ff		;

	reg									tx_valid_ff			;
	reg									tx_startofpacket_ff	;
	reg									tx_endofpacket_ff	;
	reg		[       DATA_BIT-1 : 0 ]	tx_data_ff			;
	reg		[                1 : 0 ]	tx_mod_ff			;
	reg									tx_err_ff			;

// =============================================================================
// FUNCTION DESCRIPTION
// =============================================================================

	assign	frame_start	= valid_i & frame_start_i & ~( |data_i[3:0] ) ;
	assign	frame_end	= valid_i & frame_end_i ;

	always @ ( posedge clk or negedge rst_n ) begin
		if ( !rst_n ) begin
			pkt_gen_state_ff	<= ST_IDLE ;
			ready_ff			<= 1'b0 ;
			tx_valid_pre_ff		<= 1'b0 ;
			tx_sop_pre_ff		<= 1'b0 ;
			tx_eop_pre_ff		<= 1'b0 ;
			tx_data_pre_ff		<= { DATA_BIT{1'b0} } ;
			des_mac_adr_ff		<= { MAC_ADR_BIT{1'b0} } ;
			src_mac_adr_ff		<= { MAC_ADR_BIT{1'b0} } ;
			tag_ff				<= { TAG_BIT{1'b0} } ;
			ether_type_ff		<= { TYPE_BIT{1'b0} } ;
			frame_num_ff		<= { FRM_NUM_BIT{1'b0} } ;
			pkt_number_ff		<= { PACKET_NUM_BIT{1'b0} } ;
			ttl_pkt_count_ff	<= { PACKET_NUM_BIT{1'b0} } ;
			master_clock_sec_ff	<= { T_STP_SEC_BIT{1'b0} } ;
			master_clock_ns_ff	<= { T_STP_NS_BIT{1'b0} } ;
			pkt_cnt_ff			<= 4'h0 ;
			packet_generate_ff	<= 1'b0 ;
		end
		else if ( srst ) begin
			pkt_gen_state_ff	<= ST_IDLE ;
			ready_ff			<= 1'b0 ;
			tx_valid_pre_ff		<= 1'b0 ;
			tx_sop_pre_ff		<= 1'b0 ;
			tx_eop_pre_ff		<= 1'b0 ;
			tx_data_pre_ff		<= { DATA_BIT{1'b0} } ;
			des_mac_adr_ff		<= { MAC_ADR_BIT{1'b0} } ;
			src_mac_adr_ff		<= { MAC_ADR_BIT{1'b0} } ;
			tag_ff				<= { TAG_BIT{1'b0} } ;
			ether_type_ff		<= { TYPE_BIT{1'b0} } ;
			frame_num_ff		<= { FRM_NUM_BIT{1'b0} } ;
			pkt_number_ff		<= { PACKET_NUM_BIT{1'b0} } ;
			ttl_pkt_count_ff	<= { PACKET_NUM_BIT{1'b0} } ;
			master_clock_sec_ff	<= { T_STP_SEC_BIT{1'b0} } ;
			master_clock_ns_ff	<= { T_STP_NS_BIT{1'b0} } ;
			pkt_cnt_ff			<= 4'h0 ;
			packet_generate_ff	<= 1'b0 ;
		end
		else begin
			case ( pkt_gen_state_ff ) 
				ST_IDLE		: begin
					if ( packet_transmit_i ) begin
						pkt_gen_state_ff	<= ST_STANDBY ;
						des_mac_adr_ff		<= des_mac_adr_i ;
						src_mac_adr_ff		<= src_mac_adr_i ;
						tag_ff				<= tag_i ;
						ether_type_ff		<= ether_type_i ;
					end
					ready_ff			<= 1'b1 ;
					packet_generate_ff	<= 1'b0 ;
					if ( ready_ff ) begin
						tx_valid_pre_ff		<= 1'b0 ;
						tx_eop_pre_ff		<= 1'b0 ;
					end
				end
				
				ST_STANDBY	: begin
					ready_ff	<= tx_ready_i ;
					if ( ready_ff ) begin
						if ( frame_start ) begin
							pkt_gen_state_ff	<= ST_ETH_HEAD ;
							ready_ff			<= 1'b0 ;
							frame_num_ff		<= frame_num_i ;
							pkt_number_ff		<= packet_number_i ;
							ttl_pkt_count_ff	<= total_packet_count_i ;
							master_clock_sec_ff	<= master_clock_sec_i ;
							master_clock_ns_ff	<= master_clock_ns_i ;
							packet_generate_ff	<= 1'b1 ;
						end
						tx_valid_pre_ff		<= 1'b0 ;
						tx_eop_pre_ff		<= 1'b0 ;
					end
				end
				
				ST_ETH_HEAD	: begin
					if ( ready_dly_ff ) begin
						case ( pkt_cnt_ff )
							// Destination address (top 2 byte)
							4'h0	: begin
								tx_valid_pre_ff		<= 1'b1 ;
								tx_sop_pre_ff		<= 1'b1 ;
								tx_data_pre_ff		<= { 16'h0000 , des_mac_adr_ff[47:32] } ;
							end
							// Destination address (low 4 byte) 
							4'h1	: begin
								tx_valid_pre_ff		<= 1'b1 ;
								tx_sop_pre_ff		<= 1'b0 ;
								tx_data_pre_ff		<= des_mac_adr_ff[31:0] ;
							end
							// Source address (top 4 byte)
							4'h2	: begin
								tx_valid_pre_ff		<= 1'b1 ;
								tx_sop_pre_ff		<= 1'b0 ;
								tx_data_pre_ff		<= src_mac_adr_ff[47:16] ;
							end
							// Source address (low 2 byte) & VLAN TAGE (top 2 byte)
							4'h3	: begin
								tx_valid_pre_ff		<= 1'b1 ;
								tx_sop_pre_ff		<= 1'b0 ;
								tx_data_pre_ff		<= { src_mac_adr_ff[15:0] , tag_ff[TAG_BIT-1:TAG_BIT/2] } ;
							end
							// VLAN TAGE (low 2 byte) & thernet Type 
							4'h4	: begin
								tx_valid_pre_ff		<= 1'b1 ;
								tx_sop_pre_ff		<= 1'b0 ;
								tx_data_pre_ff		<= { tag_ff[TAG_BIT/2-1:0] , ether_type_ff } ;
							end
							default	: begin
								tx_valid_pre_ff		<= tx_valid_pre_ff ;
								tx_sop_pre_ff		<= tx_sop_pre_ff ;
								tx_data_pre_ff		<= tx_data_pre_ff ;
							end
						endcase
						if ( pkt_cnt_ff == 4'h4 ) begin
							pkt_gen_state_ff	<= ST_PAY_HEAD ;
							pkt_cnt_ff			<= 4'h0 ;
						end
						else begin
							pkt_cnt_ff	<= pkt_cnt_ff + 1'b1 ;
						end
					end
				end
				
				ST_PAY_HEAD	: begin
					if ( ready_dly_ff ) begin
						case ( pkt_cnt_ff )
							// Frame number & Packet number (top 2 byte)
							4'h0	: begin
								tx_data_pre_ff	<= { frame_num_ff[FRM_NUM_BIT-1:0] , pkt_number_ff[PACKET_NUM_BIT-1:PACKET_NUM_BIT/2] } ;
							end
							// Packet number (low 2 byte) & Total packet count (top 2 byte) 
							4'h1	: begin
								tx_data_pre_ff	<= { pkt_number_ff[PACKET_NUM_BIT/2-1:0] , ttl_pkt_count_ff[PACKET_NUM_BIT-1:PACKET_NUM_BIT/2] } ;
							end
							// Total packet count (low 2 byte) & Time stamp (sec top 2 byte)
							4'h2	: begin
								tx_data_pre_ff	<= { ttl_pkt_count_ff[PACKET_NUM_BIT/2-1:0] , master_clock_sec_ff[T_STP_SEC_BIT-1:T_STP_SEC_BIT*2/3] } ;
							end
							// Time stamp (sec low 4 byte)
							4'h3	: begin
								tx_data_pre_ff	<= master_clock_sec_ff[T_STP_SEC_BIT*2/3-1:0] ;
							end
							// Time stamp (nano sec)
							4'h4	: begin
								tx_data_pre_ff	<= master_clock_ns_ff ;
							end
							default	: begin
								tx_data_pre_ff	<= tx_data_pre_ff ;
							end
						endcase
						if ( pkt_cnt_ff == 4'h4 ) begin
							pkt_gen_state_ff	<= ST_PAY_DATA ;
							pkt_cnt_ff			<= 4'h0 ;
							ready_ff			<= tx_ready_i ;
						end
						else begin
							pkt_cnt_ff	<= pkt_cnt_ff + 1'b1 ;
						end
					end
				end
				
				ST_PAY_DATA	: begin
					ready_ff	<= tx_ready_i ;
					if ( ready_ff ) begin
						tx_valid_pre_ff		<= valid_i ;
						tx_data_pre_ff		<= data_i ;
						tx_eop_pre_ff		<= frame_end_i ;
						if ( frame_end ) begin
							if ( packet_transmit_i ) begin
								pkt_gen_state_ff	<= ST_STANDBY ;
							end
							else begin
								pkt_gen_state_ff	<= ST_IDLE ;
							end
						end
					end
				end
				
				default		: begin
					pkt_gen_state_ff	<= ST_IDLE ;
				end
			endcase
		end
	end

	always @ ( posedge clk or negedge rst_n ) begin
		if ( !rst_n ) begin
			ready_dly_ff	<= 1'b0 ;
		end
		else if ( srst ) begin
			ready_dly_ff	<= 1'b0 ;
		end
		else begin
			ready_dly_ff	<= tx_ready_i ;
		end
	end

	always @ ( posedge clk or negedge rst_n ) begin
		if ( !rst_n ) begin
			tx_valid_lat_ff		<= 1'b0 ;
			tx_sop_lat_ff		<= 1'b0 ;
			tx_eop_lat_ff		<= 1'b0 ;
			tx_data_lat_ff		<= { DATA_BIT{1'b0} } ;
		end
		else if ( srst ) begin
			tx_valid_lat_ff		<= 1'b0 ;
			tx_sop_lat_ff		<= 1'b0 ;
			tx_eop_lat_ff		<= 1'b0 ;
			tx_data_lat_ff		<= { DATA_BIT{1'b0} } ;
		end
		else begin
			if ( ready_ff || ready_dly_ff ) begin
				tx_valid_lat_ff		<= tx_valid_pre_ff ;
				tx_sop_lat_ff		<= tx_sop_pre_ff ;
				tx_eop_lat_ff		<= tx_eop_pre_ff ;
				tx_data_lat_ff		<= tx_data_pre_ff ;
			end
		end
	end

	always @ ( posedge clk or negedge rst_n ) begin
		if ( !rst_n ) begin
			tx_valid_ff			<= 1'b0 ;
			tx_startofpacket_ff	<= 1'b0 ;
			tx_endofpacket_ff	<= 1'b0 ;
			tx_data_ff			<= { DATA_BIT{1'b0} } ;
			tx_mod_ff			<= 2'b00 ;
			tx_err_ff			<= 1'b0 ;
		end
		else if ( srst ) begin
			tx_valid_ff			<= 1'b0 ;
			tx_startofpacket_ff	<= 1'b0 ;
			tx_endofpacket_ff	<= 1'b0 ;
			tx_data_ff			<= { DATA_BIT{1'b0} } ;
			tx_mod_ff			<= 2'b00 ;
			tx_err_ff			<= 1'b0 ;
		end
		else begin
			if ( tx_ready_i ) begin
				tx_valid_ff			<= ( ready_ff || ready_dly_ff ) ?	tx_valid_pre_ff	: tx_valid_lat_ff ;
				tx_startofpacket_ff	<= ( ready_ff || ready_dly_ff ) ?	tx_sop_pre_ff	: tx_sop_lat_ff ;
				tx_endofpacket_ff	<= ( ready_ff || ready_dly_ff ) ?	tx_eop_pre_ff	: tx_eop_lat_ff ;
				tx_data_ff			<= ( ready_ff || ready_dly_ff ) ?	tx_data_pre_ff	: tx_data_lat_ff ;
				tx_mod_ff			<= 2'b00 ;
				tx_err_ff			<= 1'b0 ;
			end
		end
	end

	assign	ready_o				= ready_ff ;

	assign	tx_valid_o			= tx_valid_ff ;
	assign	tx_startofpacket_o	= tx_startofpacket_ff ;
	assign	tx_endofpacket_o	= tx_endofpacket_ff ;
	assign	tx_data_o			= tx_data_ff ;
	assign	tx_mod_o			= tx_mod_ff ;
	assign	tx_err_o			= tx_err_ff ;

	assign	packet_generate_o	= packet_generate_ff ;

endmodule
`default_nettype wire